/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "maya/maya.pmh"

#include "maya/MFnPlugin.h"

#define FE_MN_COMPOUND_TEST		FALSE
#define FE_MN_POPULATE_DEBUG	FALSE
#define FE_MN_CATALOG_DEBUG		FALSE
#define FE_MN_PAINT_DEBUG		FALSE
#define FE_MN_COMPUTE_DEBUG		FALSE
#define FE_MN_IO_DEBUG			FALSE

// TODO may sure this is a unique name
static const char* feMayaDefaultOutputSurface = "Output Surface";

namespace fe
{

I32 mayaDiscoverOperators(MObject a_object)
{
	MayaContext mayaContext;

#if FE_MN_POPULATE_DEBUG
	feLog("mayaDiscoverOperators master=%p\n",
		mayaContext.master().raw());
#endif

	sp<Registry> spRegistry=mayaContext.master()->registry();
	mayaContext.loadLibraries();

	MStatus status;

	MFnPlugin plugin(a_object,PLUGIN_COMPANY,"1.0","Any");

	MString feMenu;
	MGlobal::executeCommand("menu -label FE -p MayaWindow",feMenu);
	MayaNode::ms_feMenu=feMenu.asChar();

	MString feCreateMenu;
	MGlobal::executeCommand(
			("menuItem -sm true -label \"Create Operator\" -p "+
			String(feMenu.asChar())).c_str(),feCreateMenu);

	plugin.addMenuItem("Create Mesh",feMenu.asChar(),
			"createNode","mesh");

	plugin.addMenuItem("Brush",feMenu.asChar(),
			"setToolTo","MayaBrushCommon");

	status=plugin.registerContextCommand(
			"MayaBrushContext", &MayaBrush::Command::creator);
	if(!status)
	{
		status.perror("registerContextCommand");
		return status;
	}

	MayaBrush::staticSetup();

	//* TODO read from a file
	// Add SurfaceWrap: select hi-res, then driver
	String mel;
	mel+="global proc feAddSurfaceWrap()\n";
	mel+="{\n";
	mel+="	string $source[]=`ls -sl`;\n";
	mel+="	string $input=$source[0]+\".worldMesh[0]\";\n";
	mel+="	string $driver=$source[1]+\".worldMesh[0]\";\n";

	mel+="	string $freeze[]=`duplicate -name polyReference -rr $source[1]`;\n";
	mel+="	delete -ch $freeze[0];\n";
	mel+="	string $children[]=`listRelatives -children -shapes $freeze[0]`;\n";
	mel+="	string $reference=$children[0]+\".worldMesh[0]\";\n";

	mel+="	string $input_mat=$source[0]+\".matrix\";\n";
	mel+="	string $driver_mat=$source[1]+\".matrix\";\n";

	mel+="	string $xform_input=`createNode transformGeometry`;\n";
	mel+="	string $xform_input_in=$xform_input+\".inputGeometry\";\n";
	mel+="	string $xform_input_mat=$xform_input+\".transform\";\n";
	mel+="	string $xform_input_out=$xform_input+\".outputGeometry\";\n";

	mel+="	string $xform_driver=`createNode transformGeometry`;\n";
	mel+="	string $xform_driver_in=$xform_driver+\".inputGeometry\";\n";
	mel+="	string $xform_driver_mat=$xform_driver+\".transform\";\n";
	mel+="	string $xform_driver_out=$xform_driver+\".outputGeometry\";\n";

	mel+="	string $bind=`createNode FE_SurfaceBindOp`;\n";
	mel+="	string $bind_input=$bind+\".Input_Surface\";\n";
	mel+="	string $bind_driver=$bind+\".Driver_Surface\";\n";
	mel+="	string $bind_output=$bind+\".Output_Surface\";\n";
	mel+="	string $bind_prim=$bind+\".Bind_Primitives\";\n";
	mel+="	string $bind_frag=$bind+\".Bind_Fragments\";\n";

	mel+="	string $wrap=`createNode FE_SurfaceWrapOp`;\n";
	mel+="	string $wrap_input=$wrap+\".Input_Surface\";\n";
	mel+="	string $wrap_driver=$wrap+\".Driver_Surface\";\n";
	mel+="	string $wrap_deformed=$wrap+\".Deformed_Surface\";\n";
	mel+="	string $wrap_output=$wrap+\".Output_Surface\";\n";

	mel+="	string $mesh=`createNode -name polyWrapShape mesh`;\n";
	mel+="	string $children[]=`listRelatives -children -shapes $mesh`;\n";
	mel+="	string $result=$children[0]+\".inMesh\";\n";

	mel+="	setAttr $bind_prim 0;\n";
	mel+="	setAttr $bind_frag 0;\n";

	mel+="	connectAttr $input $xform_input_in;\n";
	mel+="	connectAttr $input_mat $xform_input_mat;\n";

	mel+="	connectAttr $driver $xform_driver_in;\n";
	mel+="	connectAttr $driver_mat $xform_driver_mat;\n";

	mel+="	connectAttr $xform_input_out $bind_input;\n";
	mel+="	connectAttr $reference $bind_driver;\n";

	mel+="	connectAttr $bind_output $wrap_input;\n";
	mel+="	connectAttr $reference $wrap_driver;\n";
	mel+="	connectAttr $xform_driver_out $wrap_deformed;\n";

	mel+="	connectAttr $wrap_output $result;\n";
	mel+="}\n";

	MString feMelResult;
	MGlobal::executeCommand(mel.c_str(),feMelResult);

	plugin.addMenuItem("Add SurfaceWrap",feMenu.asChar(),
			"feAddSurfaceWrap","");

	const char* exposure=getenv("FE_EXPOSURE");
	const BWORD showPrototype=exposure &&
			(exposure==String("prototype") || exposure==String("all"));

	int count=0;

	std::vector<String> available;
	spRegistry->listAvailable("OperatorSurfaceI",available);
	U32 candidates=available.size();
	for(U32 index=0;index<candidates;index++)
	{
#if FE_MN_POPULATE_DEBUG
		feLog("checking %d/%d \"%s\"\n",
				index,candidates,available[index].c_str());
#endif

		//* ignore operators tagged as prototype
		if(!showPrototype && available[index].dotMatch("*.*.*.prototype"))
		{
#if FE_CODEGEN<=FE_DEBUG
			feLog("[MayaNode] mayaDiscoverOperators"
					" hiding prototype operator:\n"
					"  %s\n",available[index].c_str());
#endif
			continue;
		}

		sp<OperatorSurfaceI> spOperatorSurfaceI=
				spRegistry->create(available[index]);
		if(!spOperatorSurfaceI.isValid())
		{
			feLog("mayaDiscoverOperators"
					" Component is not an OperatorSurfaceI \"%s\"\n",
					available[index].c_str());
			continue;
		}

		String implementation=available[index];
		char* buffer1=strchr(const_cast<char*>(available[index].c_str()),'.');
		if(buffer1)
		{
			buffer1++;
			const char* buffer2=strchr(buffer1,'.');
			if(buffer2>buffer1)
			{
				implementation.sPrintf("%.*s",int(buffer2-buffer1),buffer1);
			}
			else
			{
				implementation=buffer1;
			}
		}

#if FE_MN_POPULATE_DEBUG
		feLog("mayaDiscoverOperators implementation \"%s\"\n",
				implementation.c_str());
#endif

		const U32 uniqueId=0x80013+count;	//* HACK

		const char* name=strdup(implementation.c_str());
		const char* english=strdup(("FE_"+implementation).c_str());

		fe::MayaNode::preRegister(uniqueId,english,available[index]);
		MStatus status=plugin.registerNode(
				english,
				MTypeId(uniqueId),
				fe::MayaNode::create,
				fe::MayaNode::initialize);
		if(!status)
		{
			status.perror("registerNode");
		}
		fe::MayaNode::postRegister();

		plugin.addMenuItem(
				implementation.c_str(),
				feCreateMenu.asChar(),
				"createNode",
				english);

		count++;
	}

	return count;
}

} // namespace fe

extern "C"
{

//* discovery point for Maya
FE_DL_EXPORT MStatus initializePlugin(MObject a_object)
{
#if FE_MN_POPULATE_DEBUG
	feLog("fexMayaDL ::initializePlugin\n");
#endif

	I32 count=fe::mayaDiscoverOperators(a_object);

#if FE_MN_POPULATE_DEBUG
	feLog("fexMayaDL ::initializePlugin done\n");
#endif

	return MS::kSuccess;
}

FE_DL_EXPORT MStatus uninitializePlugin(MObject a_object)
{
#if FE_MN_POPULATE_DEBUG
	feLog("fexMayaDL ::uninitializePlugin\n");
#endif

	MGlobal::executeCommand(
			("deleteUI -m "+fe::MayaNode::ms_feMenu).c_str());

	MFnPlugin plugin(a_object);

	//* TODO proper
/*
	MStatus status=plugin.deregisterNode(fe::MayaNode::ms_id);
	if(!status)
	{
		status.perror("deregisterNode");
	}
*/

#if FE_MN_POPULATE_DEBUG
	feLog("fexMayaDL ::uninitializePlugin done\n");
#endif

	return MS::kSuccess;
}

//* TODO Is this unneeded in Maya?
#if FALSE
FE_DL_EXPORT void HoudiniDSOInit(UT_DSOInfo& dsoinfo)
{
	//* force Houdini to load with RTLD_GLOBAL (needed for dynamic_cast<>)
	dsoinfo.loadGlobal=true;
}
#endif

}

namespace fe
{

I32	MayaNode::ms_registerLock=0;
String MayaNode::ms_registerName;
String MayaNode::ms_feMenu;
MObject	MayaNode::ms_compoundBehavior;
std::map<String,MayaNode::Registration> MayaNode::ms_mayaRegistry;
std::map<String,MayaNode::Brush> MayaNode::ms_brushMap;

MayaNode::MayaNode(void):
	m_paintSize(0),
	m_bound(FALSE)
{
//	feLog("MayaNode::MayaNode master=%p\n",m_mayaContext.master().raw());

	sp<Registry> spRegistry=m_mayaContext.master()->registry();
	m_mayaContext.loadLibraries();

	m_spDrawOutput=spRegistry->create("DrawI.MayaDraw");
//	feLog("MayaNode::MayaNode m_spDrawOutput valid=%d\n",
//			m_spDrawOutput.isValid());

	//* NOTE Houdini guides are not direct to OpenGL like this
	m_spDrawGuideChain=spRegistry->create("DrawI.DrawOpenGL");

	setupScope(m_mayaContext.scope());

#if FE_COUNTED_TRACK
	Counted::registerRegion(this,sizeof(MayaNode),"MayaNode");
#endif

//	feLog("MayaNode::MayaNode done\n");
}

void MayaNode::postConstructor(void)
{
	precompute(0.0);
	updateComponents(TRUE);
}

MayaNode::~MayaNode(void)
{
//	feLog("~MayaNode\n");

	clearReferences();

#if FE_COUNTED_TRACK
	Counted::deregisterRegion(this);
#endif
}

void MayaNode::preRegister(U32 a_uniqueId,String a_name,
		String a_implementation)
{
#if FE_MN_POPULATE_DEBUG
	feLog("MayaNode::preRegister\n");
#endif

	FEASSERT(ms_registerLock==0);
	ms_registerLock++;
	ms_registerName=a_name;

	Registration& rRegistration=ms_mayaRegistry[a_name];
	rRegistration.m_autoOutput=FALSE;
	rRegistration.m_uniqueId=a_uniqueId;
	rRegistration.m_implementation=a_implementation;
}

void MayaNode::postRegister(void)
{
#if FE_MN_POPULATE_DEBUG
	feLog("MayaNode::postRegister\n");
#endif

	FEASSERT(ms_registerLock==1);

	ms_registerName="";
	ms_registerLock--;
}

MStatus MayaNode::initialize(void)
{
#if FE_MN_POPULATE_DEBUG
	feLog("MayaNode::initialize %s %d\n",
			ms_registerName.c_str(),ms_registerLock);
#endif

	populateOperator(ms_registerName);

#if FE_MN_POPULATE_DEBUG
	feLog("MayaNode::initialize done\n");
#endif

	return MS::kSuccess;
}

void* MayaNode::create(void)
{
//	feLog("MayaNode::create\n");
	return new MayaNode();
}

void MayaNode::populateOperator(String a_name)
{
	MayaContext mayaContext;

#if FE_MN_POPULATE_DEBUG
	feLog("MayaNode:populateOperator master=%p\n",
		mayaContext.master().raw());
#endif

	sp<Registry> spRegistry=mayaContext.master()->registry();

	Registration& rRegistration=ms_mayaRegistry[a_name];
	String implementation=rRegistration.m_implementation;
	std::map<String,Parameter>& rParameterMap=rRegistration.m_parameterMap;

#if FE_MN_POPULATE_DEBUG
	feLog("registration \"%s\" \"%s\"\n",
			a_name.c_str(),implementation.c_str());
#endif

	sp<OperatorSurfaceI> spOperatorSurfaceI=
			spRegistry->create(implementation);
	if(!spOperatorSurfaceI.isValid())
	{
		feLog("MayaNode:populateOperator"
				" Component is not an OperatorSurfaceI \"%s\"\n",
				implementation.c_str());
		return;
	}

	//* tried to move built-in attributes to a compound
	MStatus status=MS::kSuccess;
#if FE_MN_COMPOUND_TEST
//	MFnAttribute fnCaching;
//	fnCaching.setObject(caching);
//	fnCaching.setHidden(true);

//	MFnAttribute fnState;
//	fnState.setObject(state);
//	fnState.setHidden(true);

	MFnCompoundAttribute fnBehavior;
	ms_compoundBehavior=fnBehavior.create("NodeBehavior","behavior",&status);
	status=fnBehavior.setNiceNameOverride("Node Behavior");
	FEASSERT(status);
//	status=fnBehavior.addChild(caching);
	FEASSERT(status);
//	status=fnBehavior.addChild(state);
	FEASSERT(status);
#endif

	std::vector<String> keys;

	sp<Catalog> spCatalog=spOperatorSurfaceI;
	if(!spCatalog.isValid())
	{
		feLog("MayaNode:populateOperator"
				" Component is not a Catalog \"%s\"\n",
				implementation.c_str());

		//* TODO keep this from crashing later
	}
	else
	{
		spCatalog->catalogKeys(keys);
	}

	U32 number=keys.size();
#if FE_MN_POPULATE_DEBUG
	feLog("MayaNode::populateOperator keys %d\n",number);
#endif

	U32 outputs=0;
	for(U32 m=0;m<number;m++)
	{
		Instance instance;
		if(!spCatalog->catalogLookup(keys[m],instance))
		{
			continue;
		}
		if(instance.is< sp<Component> >())
		{
			char* keyname=strdup(keys[m].c_str());
			String io=spCatalog->catalogOrDefault<String>(keyname,"IO","input");
			if(io=="output")
			{
				outputs++;
			}
		}
	}

	if(!outputs)
	{
#if FE_MN_POPULATE_DEBUG
		feLog("MayaNode::populateOperator no output surface specified\n");
#endif

		//* refresh key list
		keys.clear();
		rRegistration.m_autoOutput=TRUE;

		if(spCatalog.isValid())
		{

			spCatalog->catalog< sp<Component> >(feMayaDefaultOutputSurface);
			spCatalog->catalog<String>(feMayaDefaultOutputSurface,"IO")=
					"output";

			spCatalog->catalogKeys(keys);
			number=keys.size();
		}
	}

	//* state
	if(true)
	{
		const char* stateName="fe_ascii";
		MFnStringData fnStringData;
		MObject defaultString=fnStringData.create(strdup(""));
		MFnTypedAttribute stateAttr;
		rParameterMap[stateName].m_object=stateAttr.create(stateName,stateName,
				MFnData::kString,defaultString);
		stateAttr.setKeyable(false);
		stateAttr.setStorable(true);
		stateAttr.setHidden(true);
		status=addAttribute(rParameterMap[stateName].m_object);
		FEASSERT(status);
	}

	for(U32 m=0;m<number;m++)
	{
#if FE_MN_POPULATE_DEBUG
		feLog("MayaNode::populateOperator key %d: \"%s\"\n",
				m,keys[m].c_str());
#endif

		if(keys[m]=="icon" || keys[m]=="url")
		{
			spCatalog->catalog<bool>(keys[m],"internal")=true;
		}
		if(spCatalog->catalogOrDefault<bool>(keys[m],"internal",false))
		{
			continue;
		}
		if(spCatalog->catalogOrDefault<bool>(keys[m],"hidden",false))
		{
			continue;
		}

		Instance instance;
		bool success=spCatalog->catalogLookup(keys[m],instance);
		if(!success)
		{
			feLog("MayaNode::populateOperator invalid Instance\n");
			continue;
		}

		char* cleanname=strdup(keys[m].substitute(" ","_").c_str());
		char* keyname=strdup(keys[m].c_str());
		String io=spCatalog->catalogOrDefault<String>(keyname,"IO","input");

		if(instance.is< sp<Component> >())
		{
			//* see meshOpNode.cpp

			String implementation=
					spCatalog->catalogOrDefault<String>(keyname,
					"implementation",
					"SurfaceAccessibleI.SurfaceAccessibleMaya");

			if(implementation=="SurfaceAccessibleI" || implementation==
					"SurfaceAccessibleI.SurfaceAccessibleMaya")
			{
				MFnTypedAttribute typedAttr;
				rParameterMap[keyname].m_object=typedAttr.create(
						cleanname,cleanname,MFnMeshData::kMesh);
				typedAttr.setNiceNameOverride(keyname);
				if(io=="input")
				{
					U32 channel=rRegistration.m_inputChannelArray.size();
					rRegistration.m_inputChannelArray.resize(channel+1);
					rRegistration.m_inputChannelArray[channel]=keyname;

					typedAttr.setStorable(true);
					typedAttr.setHidden(true);
				}
				else
				{
					U32 channel=rRegistration.m_outputChannelArray.size();
					rRegistration.m_outputChannelArray.resize(channel+1);
					rRegistration.m_outputChannelArray[channel]=keyname;

					typedAttr.setStorable(false);
					typedAttr.setWritable(false);
				}
				MStatus status=addAttribute(rParameterMap[keyname].m_object);
			}
			else if(implementation=="RampI")
			{
				rParameterMap[keyname].m_object=
						MRampAttribute::createCurveRamp(cleanname,cleanname);
				MStatus status=addAttribute(rParameterMap[keyname].m_object);
			}
			else if(implementation=="DrawI")
			{
				addBrushName(a_name);
			}
		}
		else
		{
			MFnAttribute attribute;

			if(instance.is<String>())
			{
				const char* value=strdup(instance.cast<String>().c_str());

				if(!spCatalog->catalogOrDefault<String>(
						keyname,"choice:0","").empty())
				{
					short defaultValue=0;

					U32 choiceCount=0;
					while(TRUE)
					{
						String keyChoice;
						keyChoice.sPrintf("choice:%d",choiceCount);
						const String nameChoice=
								spCatalog->catalogOrDefault<String>(
								keyname,keyChoice,"");
						if(nameChoice.empty())
						{
							break;
						}
						if(nameChoice==value)
						{
							defaultValue=choiceCount;
						}
						choiceCount++;
					}

					MFnEnumAttribute enumAttr;
					rParameterMap[keyname].m_object=
							enumAttr.create(cleanname,cleanname,defaultValue);
					attribute.setObject(enumAttr.object());

					for(U32 choice=0;choice<choiceCount;choice++)
					{
						String keyChoice;
						keyChoice.sPrintf("choice:%d",choice);
						const String nameChoice=
								spCatalog->catalogOrDefault<String>(
								keyname,keyChoice,"<error>");
						keyChoice.sPrintf("label:%d",choice);
						const String labelString=
								spCatalog->catalogOrDefault<String>(
								keyname,keyChoice,nameChoice);
//						const char* cleanChoice=strdup(
//								nameChoice.substitute(" ","_").c_str());
#if	FE_MN_POPULATE_DEBUG
						feLog("  choice %d/%d \"%s\" \"%s\"\n",
								choice,choiceCount,cleanChoice,labelChoice);
#endif
						MString fieldString=labelString.c_str();
						enumAttr.addField(fieldString,choice);
					}
				}
				else
				{
					MFnStringData fnStringData;
					MObject defaultString=fnStringData.create(value);
					MFnTypedAttribute stringAttr;
					rParameterMap[keyname].m_object=
							stringAttr.create(cleanname,cleanname,
							MFnData::kString,defaultString);
					attribute.setObject(stringAttr.object());
				}
			}
			else
			{
				MFnNumericAttribute numAttr;

				if(instance.is<bool>())
				{
					bool value=instance.cast<bool>();

					rParameterMap[keyname].m_object=
							numAttr.create(cleanname,cleanname,
							MFnNumericData::kBoolean,value);
				}
				else if(instance.is<Real>())
				{
					Real value=instance.cast<Real>();

					Real min=spCatalog->catalogOrDefault<Real>(keyname,
							"min",0.0);
					Real max=spCatalog->catalogOrDefault<Real>(keyname,
							"max",1.0);
					Real low=spCatalog->catalogOrDefault<Real>(keyname,
							"low",min);
					Real high=spCatalog->catalogOrDefault<Real>(keyname,
							"high",max);

					rParameterMap[keyname].m_object=
							numAttr.create(cleanname,cleanname,
							MFnNumericData::kDouble,value);
					numAttr.setSoftMin(low);
					numAttr.setSoftMax(high);
					numAttr.setMin(min);
					numAttr.setMax(max);

					const BWORD paintable=spCatalog->catalogOrDefault<bool>(
							keyname,"paintable",false);
					if(paintable)
					{
						const String paintableName=String(cleanname)+"Painting";
						const String paintListName=paintableName+"List";
						char* paintname=strdup(paintableName.c_str());
						char* listname=strdup(paintListName.c_str());

#if FALSE
						MFnDoubleArrayData doubleArrayData;
						MObject doubleArrayObject=
								doubleArrayData.create(MDoubleArray(30,1.0));

						MFnTypedAttribute paintAttr;
						MObject objectPaint=
								paintAttr.create(paintname,paintname,
								MFnData::kDoubleArray,doubleArrayObject);
#else
						MFnTypedAttribute paintAttr;
						MObject objectPaint=
								paintAttr.create(paintname,paintname,
								MFnData::kDoubleArray);
#endif
						paintAttr.setHidden(true);
						paintAttr.setStorable(true);
//						paintAttr.setUsesArrayDataBuilder(true);

						MFnCompoundAttribute compoundAttr;
						MObject objectPaintList=
								compoundAttr.create(listname,listname);
						compoundAttr.addChild(objectPaint);
						compoundAttr.setHidden(true);
						compoundAttr.setStorable(true);
//						compoundAttr.setArray(true);
//						compoundAttr.setUsesArrayDataBuilder(true);

						MStatus status=addAttribute(objectPaintList);

						rParameterMap[keyname].m_objectPaint=objectPaint;
						rParameterMap[keyname].m_objectPaintList=
								objectPaintList;

						String command="makePaintable"
								" -attrType doubleArray"
								" "+a_name+" "+paintableName;
#if FE_MN_POPULATE_DEBUG
						feLog("MayaNode::populateOperator MEL \"%s\"\n",
								command.c_str());
#endif

						//* making the call causes paint changes to be ignored
//						MGlobal::executeCommand(MString(command.c_str()));
					}
				}
				else if(instance.is<I32>())
				{
					I32 value=instance.cast<I32>();

					I32 min=spCatalog->catalogOrDefault<I32>(keyname,
							"min",0);
					I32 max=spCatalog->catalogOrDefault<I32>(keyname,
							"max",10);
					I32 low=spCatalog->catalogOrDefault<I32>(keyname,
							"low",min);
					I32 high=spCatalog->catalogOrDefault<I32>(keyname,
							"high",max);

					rParameterMap[keyname].m_object=
							numAttr.create(cleanname,cleanname,
							MFnNumericData::kInt,value);
					numAttr.setSoftMin(low);
					numAttr.setSoftMax(high);
					numAttr.setMin(min);
					numAttr.setMax(max);
				}
				else if(instance.is<SpatialVector>())
				{
					SpatialVector& value=instance.cast<SpatialVector>();

					rParameterMap[keyname].m_object=
							numAttr.create(cleanname,cleanname,
							MFnNumericData::k3Double);
					numAttr.setDefault((double)value[0],(double)value[1],
							(double)value[2]);
				}
				else if(instance.is<Vector4>())
				{
					Vector4& value=instance.cast<Vector4>();

					rParameterMap[keyname].m_object=
							numAttr.create(cleanname,cleanname,
							MFnNumericData::k4Double);
					numAttr.setDefault((double)value[0],(double)value[1],
							(double)value[2],(double)value[3]);
				}

				attribute.setObject(numAttr.object());
			}

			attribute.setNiceNameOverride(keyname);
			if(io=="output")
			{
				attribute.setKeyable(false);
				attribute.setStorable(false);
			}
			else
			{
				attribute.setKeyable(true);
				attribute.setStorable(true);
			}

#if FE_MN_COMPOUND_TEST
			MStatus status=
					fnBehavior.addChild(rParameterMap[keyname].m_object);
#else
			MStatus status=addAttribute(rParameterMap[keyname].m_object);
#endif
		}
	}

#if FE_MN_COMPOUND_TEST
	status=addAttribute(ms_compoundBehavior);
	FEASSERT(status);
#endif

	//* Currently every output is dependent on every input.
	//* We could get more specific.
	std::map<String,Parameter>::const_iterator it=rParameterMap.begin();
	while(it!=rParameterMap.end())
	{
		const String& key=it->first;
		String io=spCatalog->catalogOrDefault<String>(key,"IO","input");
		if(io=="output")
		{
			const Parameter& parameter=it->second;
			const MObject& rParamOutput=parameter.m_object;

			std::map<String,Parameter>::const_iterator it2=
					rParameterMap.begin();
			while(it2!=rParameterMap.end())
			{
				const String& key2=it2->first;
				String io2=
						spCatalog->catalogOrDefault<String>(key2,"IO","input");
				if(io2!="output")
				{
					const Parameter& parameter2=it2->second;
					const MObject& rParamInput=parameter2.m_object;

					MStatus status=attributeAffects(rParamInput,rParamOutput);
					if(!status)
					{
						feLog("MayaNode::populateOperator"
								" attributeAffects %s -> %s failed\n",
								key2.c_str(),key.c_str());
					}

					const MObject& rParamPaint=parameter2.m_objectPaint;
					if(!rParamPaint.isNull())
					{
						status=attributeAffects(rParamPaint,rParamOutput);
						if(!status)
						{
							feLog("MayaNode::populateOperator"
									" attributeAffects %s (Paint)"
									" -> %s failed\n",
									key2.c_str(),key.c_str());
						}
					}

					const MObject& rParamPaintList=parameter2.m_objectPaintList;
					if(!rParamPaintList.isNull())
					{
						status=attributeAffects(rParamPaintList,rParamOutput);
						if(!status)
						{
							feLog("MayaNode::populateOperator"
									" attributeAffects %s (PaintList)"
									" -> %s failed\n",
									key2.c_str(),key.c_str());
						}
					}
				}
				it2++;
			}
		}
		it++;
	}

#if FE_MN_POPULATE_DEBUG
	feLog("MayaNode::populateOperator done\n");
#endif
}

void MayaNode::updateCatalog(MDataBlock& rData,Real time,BWORD a_doInput)
{
#if FE_MN_CATALOG_DEBUG
	feLog("MayaNode::updateCatalog doInput=%d\n",a_doInput);
#endif

	if(!m_spOperatorSurfaceI.isValid())
	{
		feLog("MayaNode::updateCatalog invalid m_spOperatorSurfaceI\n");
		return;
	}

	sp<Catalog> spCatalog=m_spOperatorSurfaceI;
	if(!spCatalog.isValid())
	{
		feLog("MayaNode:updateCatalog operator is not a Catalog\n");
		return;
	}

	Registration& rRegistration=ms_mayaRegistry[m_typeName];
	const std::map<String,Parameter>& rParameterMap=
			rRegistration.m_parameterMap;

	for(std::map<String,Parameter>::const_iterator it=rParameterMap.begin();
			it!=rParameterMap.end();it++)
	{
		const String& key=it->first;
		const Parameter& parameter=it->second;
		const MObject& rObject=parameter.m_object;
		const MObject& rObjectPaint=parameter.m_objectPaint;
		const MObject& rObjectPaintList=parameter.m_objectPaintList;

#if FE_MN_CATALOG_DEBUG
		feLog("MayaNode::updateCatalog \"%s\"\n",key.c_str());
#endif

		if(key=="fe_ascii")
		{
			continue;
		}

		Instance instance;
		bool success=spCatalog->catalogLookup(key,instance);
		if(!success)
		{
			feLog("MayaNode::updateCatalog invalid Instance \"%s\"\n",
					key.c_str());
			continue;
		}

		String io=spCatalog->catalogOrDefault<String>(key,"IO","input");
		const BWORD isInput=(io=="input");
		if(io!="input output" && isInput!=a_doInput)
		{
//			feLog("MayaNode::updateCatalog skip \"%s\"\n",io.c_str());
			continue;
		}

		if(instance.is<String>())
		{
			MStatus status;
			if(a_doInput)
			{
				MDataHandle handleIn=rData.inputValue(rObject,&status);
				if(status==MS::kSuccess)
				{
					if(handleIn.type()==MFnData::kString)
					{
						spCatalog->catalog<String>(key)=
								handleIn.asString().asChar();
					}
					else
					{
						String keyChoice;
						keyChoice.sPrintf("choice:%d",handleIn.asShort());
						spCatalog->catalog<String>(key)=
							spCatalog->catalogOrDefault<String>(
							key,keyChoice,"<error>");
					}
				}
			}
			else
			{
				MDataHandle handleOut=rData.outputValue(rObject,&status);
				if(status==MS::kSuccess)
				{
					const String value=spCatalog->catalog<String>(key);

					if(handleOut.type()==MFnData::kString)
					{
						handleOut.set(MString(value.c_str()));
					}
					else
					{
						//* TODO keep a map
						U32 choiceCount=0;
						while(TRUE)
						{
							String keyChoice;
							keyChoice.sPrintf("choice:%d",choiceCount);
							const String nameChoice=
									spCatalog->catalogOrDefault<String>(
									key,keyChoice,"");
							if(nameChoice.empty())
							{
								break;
							}
							if(nameChoice==value)
							{
								handleOut.set(short(choiceCount));
								break;
							}
							choiceCount++;
						}
					}
				}
			}
		}
		else if(instance.is<bool>())
		{
			bool& rValue=spCatalog->catalog<bool>(key);

			MStatus status;
			if(a_doInput)
			{
				MDataHandle handleIn=rData.inputValue(rObject,&status);
				if(status==MS::kSuccess)
				{
					rValue=handleIn.asBool();
				}
			}
			else
			{
				MDataHandle handleOut=rData.outputValue(rObject,&status);
				handleOut.set(rValue);
			}
#if FE_MN_CATALOG_DEBUG
			feLog("MayaNode::updateCatalog now bool %d\n",rValue);
#endif

		}
		else if(instance.is<Real>())
		{
			Real& rValue=spCatalog->catalog<Real>(key);

			MStatus status;

			if(a_doInput)
			{
				MDataHandle handleIn=rData.inputValue(rObject,&status);
				if(status==MS::kSuccess)
				{
					rValue=handleIn.asDouble();
				}
			}
			else
			{
				MDataHandle handleOut=rData.outputValue(rObject,&status);
				handleOut.set(double(rValue));
			}

#if FE_MN_CATALOG_DEBUG
			feLog("MayaNode::updateCatalog now Real %.6G\n",rValue);
#endif

			const BWORD paintable=spCatalog->catalogOrDefault<bool>(
					key,"paintable",false);

#if FE_MN_CATALOG_DEBUG
			feLog("MayaNode::updateCatalog paintable %d\n",paintable);
#endif

			if(paintable)
			{
				String rPaintableName=spCatalog->catalog<String>(
						key,"paintableName");
				if(rPaintableName.empty())
				{
					rPaintableName=key.substitute(" ","_")+"Painting";
				}

				std::vector<Real>& rPainting=
						spCatalog->catalog< std::vector<Real> >(key,"painting");

				if(a_doInput)
				{
#if FE_MN_PAINT_DEBUG
					feLog("MayaNode::updateCatalog"
							" input painting size %d ---- %s\n",
							m_paintSize,m_nodeName.c_str());
#endif

#if FALSE
					MArrayDataHandle listIn=
							rData.inputArrayValue(rObjectPaintList,&status);
					if(status!=MS::kSuccess)
					{
						feLog("MayaNode::updateCatalog input failed to"
								" get handle to painting list %s\n",
								status.errorString().asChar());
					}

					U32 listCount=listIn.elementCount(&status);
					feLog("MayaNode::updateCatalog input painting"
							" listCount %d\n",listCount);
#else
					MDataHandle listIn=
							rData.inputValue(rObjectPaintList,&status);
					if(status!=MS::kSuccess)
					{
						feLog("MayaNode::updateCatalog input failed to"
								" get handle to painting list %s\n",
								status.errorString().asChar());
					}

					U32 listCount=1;
//					feLog("MayaNode::updateCatalog input painting"
//							" listCount %d\n",listCount);
#endif

					for(U32 m=0;m<listCount;m++)
					{
#if FALSE
						status=listIn.jumpToArrayElement(m);
						if(status!=MS::kSuccess)
						{
							feLog("MayaNode::updateCatalog input painting"
									" failed to jump to %d/%d %s\n",
									m,listCount,status.errorString().asChar());
						}
						const U32 index=listIn.elementIndex();
						MDataHandle elementHandle=listIn.inputValue(&status);
						if(status!=MS::kSuccess)
						{
							feLog("MayaNode::updateCatalog input painting"
									" failed to get handle %d/%d %s\n",
									m,listCount,status.errorString().asChar());
						}

						MDataHandle childHandle=
								elementHandle.child(rObjectPaint);
#else
						MDataHandle childHandle=
								listIn.child(rObjectPaint);
#endif
						MObject childData=childHandle.data();
						if(childData.isNull())
						{
#if FE_MN_PAINT_DEBUG
							feLog("MayaNode::updateCatalog input painting"
									" child data null\n");
#endif

							if(m_paintSize)
							{
#if FE_MN_PAINT_DEBUG
								feLog("MayaNode::updateCatalog input painting"
										" doubleArrayData create %d\n",
										m_paintSize);
#endif
								MFnDoubleArrayData doubleArrayData;
								MObject doubleArrayObject=
										doubleArrayData.create(
										MDoubleArray(m_paintSize,1.0));

								childHandle.set(doubleArrayObject);
								childData=childHandle.data();
							}
						}

						MFnDoubleArrayData doubleArrayData(childData,&status);
						if(status!=MS::kSuccess)
						{
#if FE_MN_PAINT_DEBUG
							feLog("MayaNode::updateCatalog input painting"
									" failed to bind doubleArrayData %s\n",
									status.errorString().asChar());
#endif
						}
						else
						{
							U32 length=doubleArrayData.length(&status);
							if(status!=MS::kSuccess)
							{
								feLog("MayaNode::updateCatalog input painting"
										" failed to get doubleArrayData"
										" length %s\n",
										status.errorString().asChar());
							}
#if FE_MN_PAINT_DEBUG
							feLog("MayaNode::updateCatalog input painting"
									" doubleArrayData length %d\n",length);
#endif

							if(!m_paintSize && length)
							{
#if FE_MN_PAINT_DEBUG
								feLog("MayaNode::updateCatalog input painting"
										" doubleArrayData reload %d\n",length);
#endif
								m_paintSize=length;
							}

							const U32 size=rPainting.size();
							if(size!=length)
							{
								rPainting.resize(length);
							}

							for(U32 m=0;m<length;m++)
							{
								rPainting[m]=doubleArrayData[m];
							}
						}
					}
				}
			}
		}
		else if(instance.is<I32>())
		{
			I32& rValue=spCatalog->catalog<I32>(key);

			MStatus status;
			if(a_doInput)
			{
				MDataHandle handleIn=rData.inputValue(rObject,&status);
				if(status==MS::kSuccess)
				{
					rValue=handleIn.asInt();
				}
			}
			else
			{
				MDataHandle handleOut=rData.outputValue(rObject,&status);
				handleOut.set(rValue);
			}
#if FE_MN_CATALOG_DEBUG
			feLog("MayaNode::updateCatalog now I32 %d\n",rValue);
#endif
		}
		else if(instance.is<SpatialVector>())
		{
			MStatus status;
			if(a_doInput)
			{
				MDataHandle handleIn=rData.inputValue(rObject,&status);
				if(status==MS::kSuccess)
				{
					double3& rValue=handleIn.asDouble3();
					set(spCatalog->catalog<SpatialVector>(key),rValue[0],
							rValue[1],rValue[2]);
				}
			}
			else
			{
				SpatialVector& rVector=spCatalog->catalog<SpatialVector>(key);

				MDataHandle handleOut=rData.outputValue(rObject,&status);
				handleOut.set(rVector[0],rVector[1],rVector[2]);
			}
		}
		else if(instance.is<Vector4>())
		{
			// TODO four floats is oddly supported
		}
		else if(instance.is< sp<Component> >())
		{
			//* meshes handled elsewhere

			sp<Component>& rspComponent=instance.cast< sp<Component> >();
			sp<MayaRamp> spRamp=rspComponent;
			if(spRamp.isValid())
			{
				MStatus status;
				MRampAttribute curveAttribute(thisMObject(),rObject,&status);
				FEASSERT(status);

				spRamp->setRamp(curveAttribute);
			}
		}
	}

	updateComponents(FALSE);

#if FE_MN_CATALOG_DEBUG
	feLog("MayaNode::updateCatalog done\n");
#endif
}

void MayaNode::updateComponents(BWORD a_constructing)
{
#if FE_MN_CATALOG_DEBUG
	feLog("MayaNode::updateComponents\n");
#endif

	sp<Catalog> spCatalog=m_spOperatorSurfaceI;
	if(!spCatalog.isValid())
	{
		feLog("MayaNode:updateComponents operator is not a Catalog\n");
		return;
	}

	Registration& rRegistration=ms_mayaRegistry[m_typeName];

	BWORD autocopyInput=FALSE;
	if(rRegistration.m_autoOutput)
	{
#if FE_MN_CATALOG_DEBUG
		feLog("MayaNode::updateComponents auto-output\n");
#endif

		spCatalog->catalog< sp<Component> >(feMayaDefaultOutputSurface);
		spCatalog->catalog<String>(feMayaDefaultOutputSurface,"IO")="output";
		autocopyInput=TRUE;
	}

	sp<Registry> spRegistry=m_mayaContext.master()->registry();

	std::vector<String> keys;
	spCatalog->catalogKeys(keys);

	const U32 number=keys.size();
	for(U32 m=0;m<number;m++)
	{
		Instance instance;
		bool success=spCatalog->catalogLookup(keys[m],instance);
		if(!success)
		{
			feLog("MayaNode::updateComponents invalid Instance \"%s\"\n",
					keys[m].c_str());
			continue;
		}

		char* cleanname=strdup(keys[m].substitute(" ","_").c_str());
		char* keyname=strdup(keys[m].c_str());

		if(instance.is< sp<Component> >())
		{
			sp<Component>& rspComponent=instance.cast< sp<Component> >();

			if(!rspComponent.isValid())
			{
				String implementation=spCatalog->catalogOrDefault<String>(
						keyname,"implementation",
						"SurfaceAccessibleI.SurfaceAccessibleMaya");
				if(!implementation.empty())
				{
					if(implementation=="SurfaceAccessibleI")
					{
						implementation=
								"SurfaceAccessibleI.SurfaceAccessibleMaya";
					}
					if(implementation=="DrawI")
					{
						implementation=(keyname==String("Brush"))?
								"DrawI.DrawCached": "DrawI.MayaDraw";
					}
					if(implementation=="RampI")
					{
						implementation="RampI.MayaRamp";
					}
					if(implementation=="OperatorGraphI")
					{
						implementation="OperatorGraphI.MayaGraph";
					}
					if(!a_constructing ||
							implementation.dotMatch("SurfaceAccessibleI"))
					{
						rspComponent=spRegistry->create(implementation);
						rspComponent->setName(keyname);
					}
				}
				sp<DrawI> spDrawI=rspComponent;
				if(spDrawI.isValid())
				{
					if(keyname==String("Brush"))
					{
						m_spDrawBrush=spDrawI;

						m_spDrawBrushOverlay=spRegistry->create(implementation);
						m_spDrawBrushOverlay->setName("BrushOverlay");

						sp<ViewI> spView=spRegistry->create("*.ViewOpenGL");
						spView->use(ViewMode::e_current);

						m_spDrawBrush->setView(spView);
						m_spDrawBrushOverlay->setView(spView);

						setupBrush(m_mayaContext.scope());

						//* NOTE not thread safe
#if FE_MN_CATALOG_DEBUG
						feLog("brush for \"%s\"\n",m_nodeName.c_str());
#endif
						ms_brushMap[m_nodeName].m_pMayaNode=this;
					}
				}
				sp<SurfaceAccessibleI> spSurfaceAccessibleI=rspComponent;
				if(spSurfaceAccessibleI.isValid())
				{
					String io=spCatalog->catalogOrDefault<String>(
							keyname,"IO","input");
					if(io=="input")
					{
						I32 channel=m_surfaceInputArray.size();
#if FE_MN_CATALOG_DEBUG
						feLog("MayaNode::updateComponents"
								" adding surface input %d \"%s\" \"%s\"\n",
								channel,cleanname,keyname);
#endif
						spCatalog->catalog<I32>(keyname,"channel")=channel;
						m_surfaceInputArray.push_back(spSurfaceAccessibleI);

						if(autocopyInput)
						{
							spCatalog->catalog<String>(
									feMayaDefaultOutputSurface,"copy")=keyname;
							autocopyInput=FALSE;
						}
					}
					else
					{
#if FE_MN_CATALOG_DEBUG
						feLog("MayaNode::updateComponents"
								" adding surface output \"%s\" \"%s\"\n",
								cleanname,keyname);
#endif
						m_spSurfaceOutput=spSurfaceAccessibleI;

						// HACK
						m_nameOutput=keyname;
					}
				}
			}
		}
	}

#if FE_MN_CATALOG_DEBUG
	feLog("MayaNode::updateComponents done\n");
#endif
}

void MayaNode::precompute(Real a_time)
{
	m_typeName=typeName().asChar();
	m_nodeName=name().asChar();

#if FE_MN_COMPUTE_DEBUG
	feLog("MayaNode::precompute \"%s\" \"%s\"\n",
			m_nodeName.c_str(),m_typeName.c_str());
#endif

	// TODO check maya node state (normal, passthrough, etc)

	Registration& rRegistration=ms_mayaRegistry[m_typeName];
	const U32 uniqueId=rRegistration.m_uniqueId;
	m_implementation=rRegistration.m_implementation;

#if FE_MN_COMPUTE_DEBUG
	feLog("  name \"%s\"\n",m_typeName.c_str());
	feLog("  unique ID %p\n",uniqueId);
	feLog("  implementation \"%s\"\n",m_implementation.c_str());
#endif

	std::map<String,Parameter>& rParameterMap=rRegistration.m_parameterMap;

	if(!m_spOperatorSurfaceI.isValid())
	{
		sp<Registry> spRegistry=m_mayaContext.master()->registry();
		m_spOperatorSurfaceI=spRegistry->create(m_implementation);

#if FE_MN_COMPUTE_DEBUG
		feLog("MayaNode::precompute m_spOperatorSurfaceI \"%s\" valid=%d\n",
				m_implementation.c_str(),m_spOperatorSurfaceI.isValid());
#endif
		m_bound=FALSE;
	}
}

MStatus MayaNode::compute(const MPlug& rPlug,MDataBlock& rData)
{
	const double time=0.0;			// TODO

	precompute(time);

	if(!m_bound)
	{
		if(m_spOperatorSurfaceI.isValid())
		{
			updateCatalog(rData,time,TRUE);

			m_lastName=m_implementation;
			insertHandler(m_spOperatorSurfaceI);
			m_bound=TRUE;

			restoreState();
		}
	}
	else
	{
		updateCatalog(rData,time,TRUE);
	}

	//* HACK icon test
//	MFnDependencyNode depNode(thisMObject());
//	depNode.setIcon("... /FE_20.png");

	Registration& rRegistration=ms_mayaRegistry[m_typeName];
	std::map<String,Parameter>& rParameterMap=rRegistration.m_parameterMap;

	if(m_spOperatorSurfaceI.isValid())
	{
		m_spOperatorSurfaceI->setName(m_nodeName);

		double frame=0;	// TODO

		sp<SurfaceAccessibleMaya> spSurfaceOutput=m_spSurfaceOutput;
		if(!spSurfaceOutput.isValid())
		{
			sp<Registry> spRegistry=m_mayaContext.master()->registry();
			spSurfaceOutput=spRegistry->create(
					"SurfaceAccessibleI.SurfaceAccessibleMaya");
			spSurfaceOutput->setName(m_nodeName+" output");
		}
#if FE_MN_COMPUTE_DEBUG
		feLog("MayaNode::compute spSurfaceOutput valid=%d\n",
				spSurfaceOutput.isValid());
#endif

		if(m_spDrawGuideCached.isValid())
		{
			m_spDrawGuideCached->clearInput();
		}

		I32 channel=0;
		if(!m_nameOutput.empty())
		{
			sp<Catalog> spCatalog=m_spOperatorSurfaceI;
			String copy=spCatalog->catalogOrDefault<String>(m_nameOutput,
					"copy","");
			channel= -1;
			if(!copy.empty())
			{
				channel=spCatalog->catalogOrDefault<I32>(copy,"channel",-1);
			}
		}

		// TODO multiple outputs with individual channels?

		String outputName=rRegistration.m_outputChannelArray[0];
		MObject paramOutMesh=rParameterMap[outputName].m_object;

		MStatus status=MS::kSuccess;
		MDataHandle outputData=rData.outputValue(paramOutMesh,&status);
		FEASSERT(status);
		MObject meshOutput=outputData.asMesh();

		MDataHandle channelData;

		U32 surfaces=m_surfaceInputArray.size();
		FEASSERT(surfaces==rRegistration.m_inputChannelArray.size());
		for(int surface=0;surface<surfaces;surface++)
		{
			sp<SurfaceAccessibleMaya> spSurfaceInput=
					m_surfaceInputArray[surface];

			String inputName=rRegistration.m_inputChannelArray[surface];
			MObject paramInMesh=rParameterMap[inputName].m_object;

			MStatus status=MS::kSuccess;

			MDataHandle inputData=rData.inputValue(paramInMesh,&status);
			FEASSERT(status);
			MObject meshInput=inputData.asMesh();

			spSurfaceInput->setMeshData(inputData);

			if(surface==channel)
			{
				channelData=inputData;
			}

#if FALSE
			// HACK alternate attribute test
			sp<SurfaceAccessorI> spAccessor=spSurfaceInput->accessor(
					SurfaceAccessibleI::e_point,"milliIndex");
			FEASSERT(spAccessor.isValid());

			if(spAccessor.isValid())
			{
				const I32 pointCount=spAccessor->count();
				for(I32 index=0;index<pointCount;index++)
				{
//					feLog("milli %d/%d\n",index,spAccessor->count());
					spAccessor->set(index,0,Real(0.001*index));
				}
			}
#endif
		}

		if(channel>=0 && channel<surfaces)
		{
#if FE_MN_COMPUTE_DEBUG
			feLog("MayaNode::compute output=input\n");
#endif

			MFnMeshData meshDataFn;
			MObject newMeshData=meshDataFn.create();

			MFnMesh newMeshFn;
			newMeshFn.copy(channelData.asMesh(),newMeshData,&status);
			if(!status)
			{
				feLog("MayaNode::compute \"%s\" mesh copy failed: %s\n",
						m_nodeName.c_str(),
						status.errorString().asChar());
				return status;
			}

			outputData.set(newMeshData);
			meshOutput=outputData.asMesh();

			const U32 vertices=newMeshFn.numVertices();
			if(m_paintSize!=vertices)
			{
				m_paintSize=vertices;

				//* second update
				updateCatalog(rData,time,TRUE);
			}
		}
		else
		{
			// erase all prior geometry

#if FE_MN_COMPUTE_DEBUG
			feLog("MayaNode::compute output reset\n");
#endif

			MFnMeshData meshDataFn;
			MObject newMeshData=meshDataFn.create();

			int numVertices=0;
			int numPolygons=0;
			MFloatPointArray vertexArray;
			MIntArray polygonCounts;
			MIntArray polygonConnects;

			status=MS::kSuccess;
			MFnMesh newMeshFn;
			MObject mesh=newMeshFn.create(numVertices,numPolygons,
					vertexArray,polygonCounts,polygonConnects,
					newMeshData,&status);
			if(!status)
			{
				feLog("MayaNode::compute create failed: %s\n",
						status.errorString().asChar());
			}

			outputData.set(newMeshData);
			meshOutput=outputData.asMesh();
			m_paintSize=0;
		}
		spSurfaceOutput->setMeshData(outputData);

		sp<MayaDraw> spMayaDraw=m_spDrawOutput;
		if(spMayaDraw.isValid())
		{
#if FE_MN_COMPUTE_DEBUG
			feLog("MayaNode::compute set draw mesh\n");
#endif
			spMayaDraw->setMeshData(outputData);
		}

		//* TODO frame, startFrame, endFrame

		m_aSurfaceOutput(m_cookSignal)=spSurfaceOutput;
		m_spSignalerI->signal(m_cookSignal);

		//* TODO draw Guide data

		updateCatalog(rData,time,FALSE);
	}

	m_lastTime=time;

/*	TODO MGlobal::
	static void			displayInfo( const MString & theMessage );
	static void			displayWarning( const MString & theWarning );
	static void			displayError( const MString & theError );
*/

	sp<Catalog> spCatalog=m_spOperatorSurfaceI;
	const MStatus status=
			spCatalog->catalogOrDefault<String>("error","").empty()?
			MS::kSuccess: MS::kFailure;

	relayErrors();

#if FE_MN_COMPUTE_DEBUG
	feLog("MayaNode::compute done\n");
#endif

	return status;
}

void MayaNode::relayErrors(void)
{
	sp<Catalog> spCatalog=m_spOperatorSurfaceI;

	String messageString=spCatalog->catalogOrDefault<String>("message","");
	if(!messageString.empty())
	{
		MGlobal::displayInfo((m_nodeName+": "+messageString).c_str());
		spCatalog->catalog<String>("message")="";

#if FE_MN_COMPUTE_DEBUG
		feLog("MayaNode::compute message string \"%s\"\n",
				messageString.c_str());
#endif
	}

	String warningString=spCatalog->catalogOrDefault<String>("warning","");
	if(!warningString.empty())
	{
		MGlobal::displayWarning((m_nodeName+": "+warningString).c_str());
		spCatalog->catalog<String>("warning")="";

#if FE_MN_COMPUTE_DEBUG
		feLog("MayaNode::compute warning string \"%s\"\n",
				warningString.c_str());
#endif
	}

	String errorString=spCatalog->catalogOrDefault<String>("error","");
	if(!errorString.empty())
	{
		MGlobal::displayError((m_nodeName+": "+errorString).c_str());
		spCatalog->catalog<String>("error")="";

#if FE_MN_COMPUTE_DEBUG
		feLog("MayaNode::compute error string \"%s\"\n",
				errorString.c_str());
#endif
	}
}

void MayaNode::brush(WindowEvent& a_rEvent,
	const SpatialVector& a_rRayOrigin,const SpatialVector& a_rRayDirection)
{
//	feLog("MayaNode::brush %s\n",c_print(a_rEvent));
//	feLog(" ray %s dir %s\n",c_print(a_rRayOrigin),c_print(a_rRayDirection));

	if(!m_spDrawBrush.isValid())
	{
		return;
	}

	m_spDrawBrush->clearInput();
	m_spDrawBrushOverlay->clearInput();

	m_aWindowEvent(m_brushSignal)=a_rEvent.record();
	m_aRayOrigin(m_brushSignal)=a_rRayOrigin;
	m_aRayDirection(m_brushSignal)=a_rRayDirection;

	m_spSignalerI->signal(m_brushSignal);
}

MStatus MayaNode::shouldSave(const MPlug& rPlug,bool& rResult)
{
	String plugName(rPlug.name().asChar());

	if(plugName.dotMatch("*.fe_ascii"))
	{
		String buffer;
		if(m_spOperatorSurfaceI.isValid() &&
				m_spOperatorSurfaceI->saveState(buffer))
		{

			Registration& rRegistration=ms_mayaRegistry[m_typeName];
			std::map<String,Parameter>& rParameterMap=
					rRegistration.m_parameterMap;
			const Parameter& parameter=rParameterMap["fe_ascii"];
			const MObject& rObject=parameter.m_object;

			MDataBlock data=forceCache();
			MStatus status;
			MDataHandle handleIn=data.inputValue(rObject,&status);
			if(status==MS::kSuccess)
			{
#if FE_MN_IO_DEBUG
				feLog("MayaNode::shouldSave \"%s\" \"%s\" -> pack state\n",
						m_typeName.c_str(),plugName.c_str());
				feLog("\"%s\"\n",buffer.c_str());
#endif

				handleIn.set(MString(buffer.c_str()));
			}
		}
	}

	MStatus status=MPxNode::shouldSave(rPlug,rResult);
	return status;
}

void MayaNode::restoreState(void)
{
	if(!m_spOperatorSurfaceI.isValid())
	{
		return;
	}

	Registration& rRegistration=ms_mayaRegistry[m_typeName];
	std::map<String,Parameter>& rParameterMap=
			rRegistration.m_parameterMap;
	const Parameter& parameter=rParameterMap["fe_ascii"];
	const MObject& rObject=parameter.m_object;

	MDataBlock data=forceCache();
	MStatus status;
	MDataHandle handleIn=data.inputValue(rObject,&status);
	if(status==MS::kSuccess)
	{
		const String buffer(handleIn.asString().asChar());
		if(!buffer.empty())
		{
#if FE_MN_IO_DEBUG
			feLog("MayaNode::restoreState \"%s\" \"%s\" -> unpack state:\n",
					m_typeName.c_str(),m_nodeName.c_str());
			feLog("\"%s\"\n",buffer.c_str());
#endif

			handleIn.set(MString(""));

			if(!m_spOperatorSurfaceI->loadState(buffer))
			{
				feLog("MayaNode::restoreState loadState failed\n");
			}
		}
	}
}

} // namespace
