/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_HAM_DEBUG		TRUE

#include "operator/operator.pmh"

using namespace fe;

void HammerOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	catalog< sp<Component> >("Brush");
	catalog<String>("Brush","implementation")="DrawI";
	catalog<String>("Brush","prompt")="Prompt here.";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Driver Surface");
	catalog<bool>("Driver Surface","optional")=true;

	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
	catalog<String>("Output Surface","copy")="Input Surface";
	catalog<bool>("Output Surface","recycle")=true;

	m_spWireframe=new DrawMode();
	m_spWireframe->setDrawStyle(DrawMode::e_wireframe);

	m_spSolid=new DrawMode();
	m_spSolid->setDrawStyle(DrawMode::e_solid);
	m_spSolid->setBackfaceCulling(FALSE);

	setIdentity(m_locator);
}

void HammerOp::handle(Record& a_rSignal)
{
	sp<DrawI> spDrawBrush;
	accessBrush(spDrawBrush,a_rSignal,e_quiet);

	sp<DrawI> spDrawOverlay;
	accessBrushOverlay(spDrawOverlay,a_rSignal,e_quiet);

#if FE_HAM_DEBUG
	feLog("HammerOp::handle brush %d brushed %d\n",
			spDrawBrush.isValid(),m_brushed);
#endif

	if(spDrawBrush.isValid())
	{
		spDrawBrush->pushDrawMode(m_spSolid);
		spDrawOverlay->pushDrawMode(m_spWireframe);

		m_brushed=TRUE;

		const Color black(0.0,0.0,0.0);
		const Color white(1.0,1.0,0.8);
		const Color grey(0.8,0.8,1.0);
		const Color red(0.8,0.0,0.0);
		const Color green(0.0,0.5,0.0);
		const Color blue(0.0,0.0,1.0);
		const Color cyan(0.0,0.8,1.0);
		const Color magenta(0.8,0.0,0.8);
		const Color yellow(1.0,1.0,0.0);
		const Color orange(1.0,0.5,0.0);
		const Color blank(0.0,0.0,0.0,0.0);

		const SpatialVector up(0.0,1.0,0.0);

		if(m_spDriver.isValid())
		{
			const SpatialVector& rRayOrigin=rayOrigin(a_rSignal);
			const SpatialVector& rRayDirection=rayDirection(a_rSignal);

			m_event.bind(windowEvent(a_rSignal));

#if FE_HAM_DEBUG
			feLog("%s\n",c_print(m_event));
#endif

			const WindowEvent::MouseButtons buttons=m_event.mouseButtons();

			SpatialVector intersection;
			set(intersection);
			SpatialVector hitNormal;
			set(hitNormal);
			BWORD impacted=FALSE;

			const Real maxDistance=100.0;
			sp<SurfaceI::ImpactI> spImpact=m_spDriver->rayImpact(
					rRayOrigin,rRayDirection,maxDistance);
			if(spImpact.isValid())
			{
				sp<SurfaceTriangles::Impact> spTriImpact=spImpact;
				if(spTriImpact.isValid())
				{
					intersection=spTriImpact->intersection();
					hitNormal=spTriImpact->normal();
#if FE_HAM_DEBUG
					const I32 face=spTriImpact->face();
					const I32 pickIndex=spTriImpact->triangleIndex();
					feLog("HammerOp::handle face %d pick %d %s\n",
							face,pickIndex,c_print(intersection));
#endif

					if(m_event.isMousePress() && buttons&WindowEvent::e_mbLeft)
					{
						setTranslation(m_locator,intersection);
					}

					drawImpact(spDrawBrush,spTriImpact,0.08,&cyan);

					impacted=TRUE;
				}
			}

			sp<ViewI> spView=spDrawOverlay->view();
			sp<ViewMode> spViewMode=spView->viewMode();
			SpatialTransform camera=spViewMode->camera();
			SpatialTransform invCamera;
			invert(invCamera,camera);

			const SpatialVector deltaX=invCamera.column(0);
			const SpatialVector deltaY=invCamera.column(1);
			const SpatialVector towardCamera=invCamera.column(2);

			manipulateTransform(spDrawBrush,m_locator,1.0,towardCamera);
		}

		spDrawOverlay->popDrawMode();
		spDrawBrush->popDrawMode();

		return;
	}

	const BWORD paramChanged=!m_brushed;
	m_brushed=FALSE;

	if(paramChanged)
	{
		if(!access(m_spDriver,"Driver Surface",e_quiet))
		{
			if(!access(m_spDriver,"Input Surface")) return;
		}
	}

	//* TODO use
	const BWORD rebuild=(paramChanged ||
			catalog<bool>("Output Surface","replaced"));

	return;
}
