/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "surface.pmh"

namespace fe
{

BWORD SurfaceAccessorBase::bindInternal(SurfaceAccessibleI::Element a_element,
		const String& a_name)
{
	m_element=a_element;

	if(!m_record.isValid())
	{
		feLog("SurfaceAccessorBase::bindInternal invalid Record\n");
		return FALSE;
	}

	SurfaceModel surfaceModelRV;
	surfaceModelRV.bind(m_record);

	sp<RecordGroup> spRG=
			surfaceModelRV.surfacePointRG();

	//* TODO multiple layouts
	RecordGroup::iterator it=spRG->begin();
	m_spRecordArray= *it;

	if(!m_spRecordArray.isValid())
	{
		feLog("SurfaceAccessorBase::bindInternal invalid RecordArray\n");
		return FALSE;
	}

	m_surfacePointRV.bind(m_spRecordArray);

	return TRUE;
}

void SurfaceAccessorBase::fragmentWith(SurfaceAccessibleI::Element a_element,
	String a_attributeName)
{
//	feLog("SurfaceAccessorBase::fragmentWith \"%s\"\n",
//			a_attributeName.c_str());

	m_fragmentMap.clear();
	if(m_spSurfaceAccessibleI.isNull())
	{
		feLog("SurfaceAccessorBase::fragmentWith"
				" m_spSurfaceAccessibleI not set\n");
		return;
	}

	//* FIX this already is the accessor
	sp<SurfaceAccessorI> spAccessor=m_spSurfaceAccessibleI->accessor(
			a_element,a_attributeName);
	if(spAccessor.isNull())
	{
		feLog("SurfaceAccessorBase::fragmentWith"
				" access to attribute \"%s\" failed\n",
				a_attributeName.c_str());
		return;
	}

	U32 elementCount=spAccessor->count();
	for(U32 elementIndex=0;elementIndex<elementCount;elementIndex++)
	{
		String fragment=spAccessor->string(elementIndex);
		Fragment& rFragment=m_fragmentMap[fragment];
		rFragment.m_indexArray.push_back(elementIndex);

//		feLog("SurfaceAccessorBase::fragmentWith %d/%d fragment \"%s\"\n",
//				elementIndex,elementCount,fragment.c_str());
	}
}

} // namespace
