/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "surface.pmh"

#define	FE_STA_DEBUG	FALSE

namespace fe
{

void SurfaceTrianglesAccessible::bind(
	sp<SurfaceAccessibleI> a_spSurfaceAccessibleI)
{
	m_spSurfaceAccessibleI=a_spSurfaceAccessibleI;
}

void SurfaceTrianglesAccessible::cache(void)
{
	//* populate surface with generic accessors

	sp<SurfaceAccessorI> spVertices=m_spSurfaceAccessibleI->accessor(
			SurfaceAccessibleI::e_primitive,SurfaceAccessibleI::e_vertices);
	sp<SurfaceAccessorI> spPointNormal=m_spSurfaceAccessibleI->accessor(
			SurfaceAccessibleI::e_point,SurfaceAccessibleI::e_normal);
	sp<SurfaceAccessorI> spPointUV=m_spSurfaceAccessibleI->accessor(
			SurfaceAccessibleI::e_point,"uv");

	const U32 primitiveCount=spVertices->count();

	U32 allocated=primitiveCount*6;
	resizeArrays(primitiveCount,allocated);

	U32 triangleIndex=0;
	U32 total=0;

	const SpatialVector zero(0.0,0.0,0.0);

	for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
		const U32 subCount=spVertices->subCount(primitiveIndex);
		if(subCount<3)
		{
			continue;
		}

#if FE_STA_DEBUG
		feLog("SurfaceTrianglesAccessible::bind %d/%d tri %d/%d verts %d\n",
				primitiveIndex,primitiveCount,total,allocated,subCount);
#endif

		FEASSERT(total<allocated);

		//* TODO
		I32 partitionIndex= -1;

		for(U32 subIndex=0;subIndex<subCount;subIndex++)
		{
			const I32 pointIndex=
					spVertices->integer(primitiveIndex,subIndex);

			m_pVertexArray[total]=
					spVertices->spatialVector(primitiveIndex,subIndex);
			m_pNormalArray[total]=(spPointNormal.isValid())?
					spPointNormal->spatialVector(pointIndex): zero;
			m_pUVArray[total]=(spPointUV.isValid())?
					spPointUV->spatialVector(pointIndex): zero;
			m_pTriangleIndexArray[total]=triangleIndex;
			m_pTriangleIndexArray[total]=triangleIndex+(subIndex==3);
			m_pPrimitiveIndexArray[total]=primitiveIndex;
			m_pPartitionIndexArray[total]=partitionIndex;

#if FE_STA_DEBUG
			feLog("  %d/%d %s\n",subIndex,subCount,
					c_print(m_pVertexArray[total]));
#endif

			total++;

			if(subIndex==3)
			{
				triangleIndex++;

				m_pVertexArray[total]=m_pVertexArray[total-4];
				m_pNormalArray[total]=m_pNormalArray[total-4];
				m_pUVArray[total]=m_pUVArray[total-4];
				m_pTriangleIndexArray[total]=triangleIndex;
				m_pPrimitiveIndexArray[total]=primitiveIndex;
				m_pPartitionIndexArray[total]=partitionIndex;
				total++;

				m_pVertexArray[total]=m_pVertexArray[total-3];
				m_pNormalArray[total]=m_pNormalArray[total-3];
				m_pUVArray[total]=m_pUVArray[total-3];
				m_pTriangleIndexArray[total]=triangleIndex;
				m_pPrimitiveIndexArray[total]=primitiveIndex;
				m_pPartitionIndexArray[total]=partitionIndex;
				total++;
			}
		}

		triangleIndex++;
	}

	if(allocated!=total)
	{
		resizeArrays(primitiveCount,total);
	}

	m_elements=triangleIndex;
	m_vertices=total;

	calcBoundingSphere();
}

} // namespace
